import functools
import inspect
from copy import copy
from typing import Dict, Type, Any, TypeVar, Callable, Set, List, Optional, Union

from .interfaces import SpecialDepDefinition, ReadableContainer, TypeResolver
from .exceptions import (
    UnresolvableType,
    DuplicateDefinition,
    InvalidDependencyDefinition,
)
from .definitions import normalise, Singleton, Construction
from .util.reflection import RETURN_ANNOTATION
from .wrapping import bound_function

UNRESOLVABLE_TYPES = [str, int, float, bool]

X = TypeVar("X")


class Container(ReadableContainer):
    """ Dependency injection container

    Lagom is a dependency injection container designed to give you "just enough"
    help with building your dependencies. The intention is that almost
    all of your code doesn't know about or rely on lagom. Lagom will
    only be involved at the top level to pull everything together.

    >>> from tests.examples import SomeClass
    >>> c = Container()
    >>> c[SomeClass]
    <tests.examples.SomeClass object at ...>

    Objects are constructed as they are needed

    >>> from tests.examples import SomeClass
    >>> c = Container()
    >>> first = c[SomeClass]
    >>> second = c[SomeClass]
    >>> first != second
    True

    And construction logic can be defined
    >>> from tests.examples import SomeClass, SomeExtendedClass
    >>> c = Container()
    >>> c[SomeClass] = SomeExtendedClass
    >>> c[SomeClass]
    <tests.examples.SomeExtendedClass object at ...>
    """

    _registered_types: Dict[Type, SpecialDepDefinition]
    _explicitly_registered_types: Set[Type]

    def __init__(self, container: Optional["Container"] = None):
        """
        :param container: Optional container if provided the existing definitions will be copied
        """
        self._registered_types = {}
        self._explicitly_registered_types = set()
        if container:
            self._registered_types = copy(container._registered_types)

    def define(self, dep: Type[X], resolver: TypeResolver[X]) -> None:
        """Register how to construct an object of type X

        >>> from tests.examples import SomeClass
        >>> c = Container()
        >>> c.define(SomeClass, lambda: SomeClass())

        :param dep: The type to be constructed
        :param resolver: A definition of how to construct it
        :return:
        """
        if dep in UNRESOLVABLE_TYPES:
            raise InvalidDependencyDefinition()
        if dep in self._explicitly_registered_types:
            raise DuplicateDefinition()
        self._registered_types[dep] = normalise(resolver)
        self._explicitly_registered_types.add(dep)

    @property
    def defined_types(self) -> Set[Type]:
        """The types the container has explicit build instructions for

        :return:
        """
        return set(self._registered_types.keys())

    def resolve(
        self, dep_type: Type[X], suppress_error=False, skip_definitions=False
    ) -> X:
        """Constructs an object of type X

         If the object can't be constructed an exception will be raised unless
         supress errors is true

        >>> from tests.examples import SomeClass
        >>> c = Container()
        >>> c.resolve(SomeClass)
        <tests.examples.SomeClass object at ...>

        >>> from tests.examples import SomeClass
        >>> c = Container()
        >>> c.resolve(int)
        Traceback (most recent call last):
        ...
        lagom.exceptions.UnresolvableType: ...

        Optional wrappers are stripped out to be what is being asked for
        >>> from tests.examples import SomeClass
        >>> c = Container()
        >>> c.resolve(Optional[SomeClass])
        <tests.examples.SomeClass object at ...>

        :param dep_type: The type of object to construct
        :param suppress_error: if true returns None on failure
        :param skip_definitions:
        :return:
        """
        try:
            optional_dep_type = _remove_optional_type(dep_type)
            if optional_dep_type:
                return self.resolve(optional_dep_type, suppress_error=True)
            if dep_type in UNRESOLVABLE_TYPES:
                raise UnresolvableType(dep_type)
            type_to_build = (
                self._registered_types.get(dep_type, dep_type)
                if not skip_definitions
                else dep_type
            )
            if isinstance(type_to_build, SpecialDepDefinition):
                return type_to_build.get_instance(self)
            return self._reflection_build(type_to_build)
        except UnresolvableType as inner_error:
            if not suppress_error:
                raise UnresolvableType(dep_type) from inner_error
            return None  # type: ignore

    def partial(
        self,
        func: Callable[..., X],
        shared: List[Type] = None,
        keys_to_skip=None,
        skip_pos_up_to: int = 0,
    ) -> Callable[..., X]:
        """Takes a callable and returns a callable bound to the container
        When invoking the new callable if any arguments can be constructed by the container
        then they can be ommited.
        >>> from tests.examples import SomeClass
        >>> c = Container()
        >>> def my_func(something: SomeClass):
        ...   return f"Successfully called with {something}"
        >>> bound_func = c.partial(my_func)
        >>> bound_func()
        'Successfully called with <tests.examples.SomeClass object at ...>'

        :param func: the function to bind to the container
        :param shared: items which should be considered singletons on a per call level
        :param keys_to_skip: named arguments which the container shouldnt build
        :param skip_pos_up_to: positional arguments which the container shouldnt build
        :return:
        """
        if shared:
            return self._partial_with_shared_singletons(func, shared)
        spec = inspect.getfullargspec(func)

        def _bind_func(extra_keys_to_skip=None, extra_skip_pos_up_to=0):
            final_keys_to_skip = (keys_to_skip or []) + (extra_keys_to_skip or [])
            final_skip_pos_up_to = max(skip_pos_up_to, extra_skip_pos_up_to)
            bindable_deps = self._infer_dependencies(
                spec,
                suppress_error=True,
                keys_to_skip=final_keys_to_skip,
                skip_pos_up_to=final_skip_pos_up_to,
            )
            return functools.partial(func, **bindable_deps)

        return bound_function(_bind_func, func)

    def clone(self):
        """ returns a copy of the container
        :return:
        """
        return Container(self)

    def __getitem__(self, dep: Type[X]) -> X:
        return self.resolve(dep)

    def __setitem__(self, dep: Type[X], resolver: TypeResolver[X]):
        self.define(dep, resolver)

    def _reflection_build(self, dep_type: Type[X]) -> X:
        spec = inspect.getfullargspec(dep_type.__init__)
        sub_deps = self._infer_dependencies(spec)
        try:
            return dep_type(**sub_deps)  # type: ignore
        except TypeError as type_error:
            raise UnresolvableType(dep_type) from type_error

    def _infer_dependencies(
        self,
        spec: inspect.FullArgSpec,
        suppress_error=False,
        keys_to_skip=None,
        skip_pos_up_to=0,
    ):
        supplied_arguments = spec.args[0:skip_pos_up_to]
        keys_to_skip = (keys_to_skip or []) + supplied_arguments
        annotations = copy(spec.annotations)
        annotations.pop(RETURN_ANNOTATION, None)
        sub_deps = {
            key: self.resolve(sub_dep_type, suppress_error=suppress_error)
            for (key, sub_dep_type) in annotations.items()
            if sub_dep_type != Any and key not in keys_to_skip
        }
        filtered_deps = {key: dep for (key, dep) in sub_deps.items() if dep is not None}
        return filtered_deps

    def _partial_with_shared_singletons(
        self, func: Callable[..., X], shared: List[Type]
    ):
        def _cloned_container():
            temp_container = self.clone()
            # For each of the shared dependencies resolve before invocation
            # and replace with a singleton
            for type_def in shared:
                temp_container[type_def] = Singleton(
                    Construction(lambda: self.resolve(type_def))
                )
            return temp_container

        def _bind_func(keys_to_skip=None, skip_pos_up_to=0):
            temp_container = _cloned_container()
            return temp_container.partial(
                func,
                shared=[],
                keys_to_skip=keys_to_skip,
                skip_pos_up_to=skip_pos_up_to,
            )

        return bound_function(_bind_func, func)


def _remove_optional_type(dep_type):
    """ if the Type is Optional[T] returns T else None

    :param dep_type:
    :return:
    """
    try:
        # Hacky: an optional type has [T, None] in __args__
        if len(dep_type.__args__) == 2 and dep_type.__args__[1] == None.__class__:
            return dep_type.__args__[0]
    except:
        pass
    return None



"""
Django integration.

Usage:


```
#views.py

from .dependency_config import container


@container.bind_view
def index(request, dep: SomeDep, questions: DjangoModel[Question]):
    new_question = questions.new(question_text="What's next?", pub_date=timezone.now())
    new_question.save()
    return HttpResponse(f"plain old function: {dep.message} with {questions.objects.all().count()} questions")


@container.bind_view
class CBVexample(View):
    def get(self, request, dep: SomeDep):
        return HttpResponse(f"Class based: {dep.message}")

```

```
# dependency_config.py
# a per app dep injection container

# Lists all models that should be available
container = DjangoContainer(models=[Question], request_singletons= [SomeCache])
container[SomeService] = SomeService("connection details etc")
```

"""
import types
from typing import TypeVar, Generic, List, Type, Optional

from django.db.models import Manager, Model
from django.views import View

from lagom import Container

M = TypeVar("M", bound=Model)


class _Managers(Generic[M]):
    """
    Wraps around a django model and provides access to the class properties.
    The intention is that all the Manager objects can be accessed via this.
    """

    def __init__(self, model: Type[M]):
        self.model = model

    def __getattr__(self, item) -> Manager:
        if not hasattr(self.model, item):
            raise KeyError(
                f"Django model {self.model.__name__} does not define a property {item}"
            )
        return getattr(self.model, item)


class DjangoModel(Generic[M]):
    """Wrapper around a django model for injection hinting
    Usage:
        container = DjangoContainer(models=[Question])

        @bind_to_container(container)
        def load_first_question(questions: DjangoModel[Question]):
            return questions.objects.first()

    """

    model: Type[M]

    def __init__(self, model: Type[M]):
        """

        :param model: The django model class
        """
        self.model = model
        self.managers = _Managers(self.model)

    @property
    def objects(self) -> Manager:
        """ Equivalent to MyModel.objects

        :return:
        """
        return self.managers.objects

    def new(self, **kwargs) -> M:
        """ Equivalent to MyModel(**kwargs)

        :param kwargs:
        :return:
        """
        return self.model(**kwargs)


class DjangoContainer(Container):
    """
    Same behaviour as the basic container bug provides a view method which
    should be used as a decorator to wrap views. Once wrapped
    the view can reference dependencies which will be auto-wired.
    """

    _request_singletons: List[Type]

    def __init__(
        self,
        models: Optional[List[Type[Model]]] = None,
        request_singletons: Optional[List[Type]] = None,
        container: Container = None,
    ):
        """
        :param models: List of models which should be available for injection
        :param request_singletons:
        :param container:
        """
        super().__init__(container)
        self._request_singletons = request_singletons or []
        for model in models or []:
            self.define(DjangoModel[model], DjangoModel(model))  # type: ignore

    def bind_view(self, view):
        """
        Takes either a plain function view or a class based view
        binds it to the container then returns something that can
        be used in a django url definition
        :param view:
        :return:
        """
        if isinstance(view, types.FunctionType):
            # Plain old function can be bound to the container
            return self.partial(view, shared=self._request_singletons)
        return self._bind_view_methods_to_container(view)

    def _bind_view_methods_to_container(self, view):
        for method in View.http_method_names:
            if hasattr(view, method):
                setattr(
                    view,
                    method,
                    self.partial(
                        getattr(view, method), shared=self._request_singletons
                    ),
                )
        return view



"""Code to help understand functions
"""
import inspect
from typing import Callable


def arity(func: Callable) -> int:
    """Returns the arity(number of args)
    Given a callable this function returns the number of arguments it expects
    >>> arity(lambda: 5)
    0
    >>> arity(lambda x: x)
    1

    :param func:
    :return:
    """
    return len(inspect.signature(func).parameters)



import inspect
from typing import TypeVar, Callable, List

INTROSPECTION_ATTRS = (
    "__module__",
    "__name__",
    "__qualname__",
    "__doc__",
    "__annotations__",
)

T = TypeVar("T")


def decorated_wrapper(wrapped: T, original) -> T:
    for attr in INTROSPECTION_ATTRS:
        if hasattr(original, attr):
            setattr(wrapped, attr, getattr(original, attr))
    return wrapped


# A function builder is wrapped around a supplied function and
# bound to the container. When called a new function will be returned
# which is the same as the original function but with some arguments supplied
# from the container. The arguments tell the container what it doesn't need to build
# first argument is a list of kwargs to skip the second the number of positional arguments
# to ignore.
FunctionBuilder = Callable[[List[str], int], Callable]


def bound_function(function_builder: FunctionBuilder, original_function: Callable):
    if inspect.iscoroutinefunction(original_function):

        async def _wrapped_function(*args, **kwargs):
            named_args_to_skip = list(kwargs.keys())
            pos_args_to_skip = len(args)
            return await function_builder(named_args_to_skip, pos_args_to_skip)(
                *args, **kwargs
            )

    else:

        def _wrapped_function(*args, **kwargs):
            named_args_to_skip = list(kwargs.keys())
            pos_args_to_skip = len(args)
            return function_builder(named_args_to_skip, pos_args_to_skip)(
                *args, **kwargs
            )

    return decorated_wrapper(_wrapped_function, original_function)



from starlette.routing import Route

from lagom import Container
from lagom.integrations.starlette import StarletteContainer


class MyDep:
    pass


class ComplexDep:
    def __init__(self, something):
        pass


def some_handler(request, dep: MyDep):
    return "ok"


def two_dep_handler(request, dep_one: MyDep, dep_two: MyDep):
    return "singleton" if dep_one is dep_two else "new instances"


def test_a_special_starlette_container_can_be_used_and_provides_routes():
    sc = StarletteContainer()
    route = sc.route("/", some_handler)
    assert isinstance(route, Route)
    assert route.endpoint({}) == "ok"


def test_the_starlette_container_can_define_request_level_singletons():
    sc = StarletteContainer(request_singletons=[MyDep])
    route = sc.route("/two", two_dep_handler)
    assert route.endpoint({}) == "singleton"


def test_the_starlette_container_can_wrap_an_existing_container(container: Container):
    container[ComplexDep] = ComplexDep("")
    sc = StarletteContainer(container=container)
    assert isinstance(sc[ComplexDep], ComplexDep)



from typing import Optional

from lagom import Container


class MySimpleDep:
    extra_stuff = "yes"


class MyComplexDep:
    extra_stuff = "yes"

    def __init__(self, something):
        pass


class MyDepWithAnOptional:
    success = "yes"

    def __init__(self, dep: Optional[MySimpleDep] = None):
        self.dep = dep


class MyDepWithAnOptionalThatCantBeBuilt:
    success = "yes"

    def __init__(self, dep: Optional[MyComplexDep] = "weird default"):  # type: ignore
        self.dep = dep


def test_missing_optional_dependencies_cause_no_errors(container: Container):
    resolved = container.resolve(MyDepWithAnOptionalThatCantBeBuilt)
    assert resolved.success == "yes"  # type: ignore


def test_defaults_for_optional_types_are_honoured(container: Container):
    resolved = container.resolve(MyDepWithAnOptionalThatCantBeBuilt)
    assert resolved.dep == "weird default"  # type: ignore


def test_optional_dependencies_are_understood_and_injected(container: Container):
    resolved = container.resolve(MyDepWithAnOptional)
    assert resolved.dep.extra_stuff == "yes"  # type: ignore


def test_we_can_ask_for_optional_things_that_cant_be_constructed(container: Container):
    assert container.resolve(Optional[MyComplexDep]) is None



import pytest

from lagom import Construction, Container, Singleton
from lagom.exceptions import DuplicateDefinition


class InitialDep:
    pass


class SomeOtherThing(InitialDep):
    pass


def test_container_can_list_the_types_explicitly_defined(container: Container):
    container[InitialDep] = InitialDep
    container[SomeOtherThing] = Singleton(SomeOtherThing)

    assert container.defined_types == {InitialDep, SomeOtherThing}



from lagom.definitions import X
from lagom.interfaces import SpecialDepDefinition, ReadableContainer


class PlainFunction(SpecialDepDefinition[X]):
    """Preserves a function without any dep injection performed on it"""

    callable_func: X

    def __init__(self, callable_func: X):
        """

        """
        self.callable_func = callable_func

    def get_instance(self, _container: ReadableContainer) -> X:
        return self.callable_func



from typing import Type, TypeVar

from lagom import Container
from lagom.experimental.exceptions import DependencyNotDefined

X = TypeVar("X")


class ExplicitContainer(Container):
    def resolve(
        self, dep_type: Type[X], suppress_error=False, skip_definitions=False
    ) -> X:
        if dep_type not in self.defined_types:
            if suppress_error:
                return None  # type: ignore
            raise DependencyNotDefined(
                f"{dep_type.__name__} has not been defined. "
                "In an explict container all dependencies must be defined"
            )
        return super().resolve(dep_type, suppress_error, skip_definitions)



"""
Classes representing specific ways of representing dependencies
"""
import inspect
from typing import Union, Type, Optional, Callable, TypeVar, Any

from .exceptions import InvalidDependencyDefinition
from .interfaces import SpecialDepDefinition, ReadableContainer
from .util.functional import arity

X = TypeVar("X")


class Construction(SpecialDepDefinition[X]):
    """Wraps a callable for constructing a type"""

    constructor: Union[Callable[[], X], Callable[[Any], X]]

    def __init__(self, constructor):
        if arity(constructor) > 1:
            raise InvalidDependencyDefinition(
                f"Arity {arity} functions are not supported"
            )
        self.constructor = constructor

    def get_instance(self, container: ReadableContainer) -> X:
        resolver = self.constructor
        artiy = arity(resolver)
        if artiy == 0:
            return resolver()  # type: ignore
        if artiy == 1:
            return resolver(container)  # type: ignore
        raise Exception("The constructor should have stopped us getting here")


class Alias(SpecialDepDefinition[X]):
    """When one class is asked for the other should be returned"""

    alias_type: Type[X]

    def __init__(self, alias_type):
        self.alias_type = alias_type

    def get_instance(self, container: ReadableContainer) -> X:
        return container.resolve(self.alias_type, skip_definitions=True)


class Singleton(SpecialDepDefinition[X]):
    """Builds only once then saves the built instance"""

    singleton_type: SpecialDepDefinition
    _instance: Optional[X]

    def __init__(self, singleton_type):
        self.singleton_type = normalise(singleton_type)
        self._instance = None

    def get_instance(self, container: ReadableContainer) -> X:
        if self._has_instance:
            return self._instance  # type: ignore
        instance = self.singleton_type.get_instance(container)
        return self._set_instance(instance)

    @property
    def _has_instance(self):
        return self._instance is not None

    def _set_instance(self, instance: X):
        self._instance = instance
        return instance


def normalise(resolver: Any) -> SpecialDepDefinition:
    """
    :param resolver:
    :return:
    """
    if isinstance(resolver, SpecialDepDefinition):
        return resolver
    elif inspect.isfunction(resolver):
        return Construction(resolver)
    elif not inspect.isclass(resolver):
        return Singleton(lambda: resolver)
    else:
        return Alias(resolver)



"""
FastAPI (https://fastapi.tiangolo.com/)

"""
import typing

from fastapi import Depends

from .. import Container

T = typing.TypeVar("T")


class FastApiContainer(Container):
    """
    Basic container plus a depends method which is drop in compatible with
    FastApi `Depends`
    """

    def depends(self, dep_type: typing.Type[T]) -> T:
        """Returns a Depends object which FastAPI understands

        :param dep_type:
        :return:
        """
        return Depends(lambda: self.resolve(dep_type))



"""
This module provides decorators for hooking an
application into the container.s
"""
import inspect
from typing import List, Type

from .definitions import Singleton
from .container import Container
from .exceptions import MissingReturnType
from .util.reflection import RETURN_ANNOTATION


def bind_to_container(container: Container, shared: List[Type] = None):
    """Decorates the function so that it's uses the container to construct things

    >>> from tests.examples import SomeClass
    >>> c = Container()
    >>> @bind_to_container(c)
    ... def say_hello_from(sayer: SomeClass):
    ...    return f"hello from {sayer}"
    >>> say_hello_from()
    'hello from <tests.examples.SomeClass object at ...>'

    """

    def _decorator(func):
        return container.partial(func, shared=shared)

    return _decorator


def dependency_definition(container: Container, singleton: bool = False):
    """ Registers the provided function with the container
    The return type of the decorated function will be reflected and whenever
    the container is asked for this type the function will be called

    >>> from tests.examples import SomeClass, SomeExtendedClass
    >>> c = Container()
    >>> @dependency_definition(c)
    ... def build_some_class_but_extended() -> SomeClass:
    ...    return SomeExtendedClass()
    >>> c.resolve(SomeClass)
    <tests.examples.SomeExtendedClass object at ...>

    """

    def _decorator(func):
        try:
            arg_spec = inspect.getfullargspec(func)
            return_type = arg_spec.annotations[RETURN_ANNOTATION]
        except KeyError:
            raise MissingReturnType(
                f"Function {func.__name__} used as a definition must have a return type"
            )
        if singleton:
            container.define(return_type, Singleton(func))
        else:
            container.define(return_type, func)
        return func

    return _decorator



from typing import Callable


from lagom import Container
from lagom.experimental.definitions import PlainFunction


class ClassNeedingAFunction:
    def __init__(self, adder: Callable[[int, int], int]):
        self.adder = adder

    def trigger(self, x, y):
        adder = self.adder
        return adder(x, y)


def add_stuff(x: int, y: int) -> int:
    return x + y


def test_ways_of_constructing_functions_can_be_provided(container: Container):
    container[Callable[[int, int], int]] = PlainFunction(add_stuff)
    erm = container[ClassNeedingAFunction]
    assert erm.trigger(2, 3) == 5



"""
Interfaces shared by modules within the lagom package
"""
from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Type, Any, Callable, Union, List, Set

X = TypeVar("X")


BuildingFunction = Callable[[Any], Any]


class ReadableContainer(ABC):
    """
    Represents a container that can resolve dependencies
    """

    @abstractmethod
    def resolve(
        self, dep_type: Type[X], suppress_error=False, skip_definitions=False
    ) -> X:
        """Constructs an object of type X
        """
        pass

    @abstractmethod
    def __getitem__(self, dep: Type[X]) -> X:
        """Shortcut to calling resolve
        """
        pass

    @property
    @abstractmethod
    def defined_types(self) -> Set[Type]:
        """Set of all the types defined in the container
        """
        pass


class SpecialDepDefinition(ABC, Generic[X]):
    """
    Represents a special way of loading a dependency.
    """

    @abstractmethod
    def get_instance(self, container: ReadableContainer) -> X:
        """ constructs the represented type(X).

        :param container: an instance of the current container
        :return:
        """
        pass


T = TypeVar("T")

"""
The TypeResolver represents the way that lagom can be
told about how to define a type. Any of these types
can be assigned to the container.
"""
TypeResolver = Union[
    Type[T],  # An alias from one type to the next
    Callable[[], T],  # A resolution function
    Callable[[ReadableContainer], T],  # A resolution function that takes the container
    SpecialDepDefinition[T],  # From the definitions module
    T,  # Just an instance of the type - A singleton
]



from flask import Flask, Response
from lagom.integrations.flask import FlaskContainer


class ComplexDep:
    def __init__(self, message):
        self.message = message


def test_flask_container_provides_a_route_decorator():
    app = Flask(__name__)
    container = FlaskContainer(app)
    container[ComplexDep] = ComplexDep("hello from dep")

    @container.route("/")
    def _some_handler(dep: ComplexDep):
        return dep.message

    with app.test_client() as client:
        resp: Response = client.get("/")
        assert resp.get_data(as_text=True) == "hello from dep"


def test_the_route_decorator_can_have_request_level_singletons():
    app = Flask(__name__)
    container = FlaskContainer(app, request_singletons=[ComplexDep])
    container[ComplexDep] = lambda: ComplexDep("hello from dep")

    @container.route("/")
    def _some_handler(dep_one: ComplexDep, dep_two: ComplexDep):
        return "singleton" if dep_one is dep_two else "nope"

    with app.test_client() as client:
        resp: Response = client.get("/")
        assert resp.get_data(as_text=True) == "singleton"



"""
Exceptions raised by the library
"""
from abc import ABC
from typing import Type

import typing


class LagomException(Exception, ABC):
    """All exceptions in this library are instances of a LagomException"""

    pass


class InvalidDependencyDefinition(ValueError, LagomException):
    """The provided construction logic is not valid"""

    pass


class MissingReturnType(SyntaxError, LagomException):
    """The function provided doesnt type hint a return"""

    pass


class DuplicateDefinition(ValueError, LagomException):
    """The type has already been defined somewhere else"""

    pass


class UnresolvableType(ValueError, LagomException):
    """The type cannot be constructed"""

    dep_type: str

    def __init__(self, dep_type: Type):
        """

        :param dep_type: The type that could not be constructed
        """
        # This first check makes 3.6 behave the same as 3.7 and later
        if hasattr(typing, "GenericMeta") and isinstance(dep_type, typing.GenericMeta):
            self.dep_type = str(dep_type)
        elif hasattr(dep_type, "__name__"):
            self.dep_type = dep_type.__name__
        else:
            self.dep_type = str(dep_type)
        super().__init__(
            f"Unable to construct dependency of type {self.dep_type} "
            "The constructor probably has some unresolvable dependencies"
        )



import pytest

from lagom import Container
from lagom.experimental.exceptions import DependencyNotDefined


class SomethingToBuild:
    pass


class SomethingBiggerToBuild:
    def __init__(self, inner: SomethingToBuild):
        pass


def test_dependencies_must_be_explicitly_defined(explicit_container: Container):
    with pytest.raises(DependencyNotDefined) as e_info:
        explicit_container.resolve(SomethingToBuild)

    assert (
        str(e_info.value)
        == "SomethingToBuild has not been defined. In an explict container all dependencies must be defined"
    )


def test_the_error_suppression_flag_is_honoured(explicit_container: Container):
    assert explicit_container.resolve(SomethingToBuild, suppress_error=True) is None


def test_an_alias_counts_as_definition(explicit_container: Container):
    explicit_container[SomethingToBuild] = SomethingToBuild

    built_object = explicit_container.resolve(SomethingToBuild)
    assert isinstance(built_object, SomethingToBuild)


def test_the_explicit_rule_applies_to_inner_objects_too(explicit_container: Container):
    explicit_container[SomethingBiggerToBuild] = SomethingBiggerToBuild

    with pytest.raises(DependencyNotDefined) as e_info:
        explicit_container.resolve(SomethingBiggerToBuild)

    assert (
        str(e_info.value)
        == "SomethingToBuild has not been defined. In an explict container all dependencies must be defined"
    )


def test_the_inner_dependencies_dont_have_to_be_defined_in_the_container(
    explicit_container: Container,
):
    explicit_container[SomethingBiggerToBuild] = lambda: SomethingBiggerToBuild(
        SomethingToBuild()
    )

    built_object = explicit_container.resolve(SomethingBiggerToBuild)
    assert isinstance(built_object, SomethingBiggerToBuild)

    # We still can't build the inner object directly
    assert explicit_container.resolve(SomethingToBuild, suppress_error=True) is None
